#ifndef NO_RCSID
const char* QN_MLPWeightFile_Tnet_rcsid =
    "$Header: /u/drspeech/repos/quicknet2/QN_MLPWeightFile_Tnet.cc,v 1.6 2014/02/03 20:49:27 davidj Exp $";
#endif

/* Must include the config.h file first */
#include <QN_config.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include "QN_types.h"
#include "QN_Logger.h"
#include "QN_utils.h"
#include "QN_libc.h"
#include "QN_MLPWeightFile_Tnet.h"
#include "QN_fltvec.h"
#include "QN_intvec.h"

// Constructor for a TNet-style weight file

QN_MLPWeightFile_Tnet::QN_MLPWeightFile_Tnet(int a_debug, 
					     const char* a_dbgname,
					     FILE* a_stream,
					     QN_FileMode a_mode,
					     size_t a_layers,
					     const size_t* a_layer_units,
					     QN_OutputLayerType a_out_layer_type,
					     size_t a_skip_sections)
    : clog(a_debug, "QN_MLPWeightFile_Tnet", a_dbgname),
      stream(a_stream),
      mode(a_mode),
      n_layers(a_layers),
      out_layer_type(a_out_layer_type),
      skip_sections(a_skip_sections)
{
    size_t i;

    assert((size_t) QN_MAX_LAYERS <= (size_t) QN_MLP_MAX_LAYERS);
    if (a_layers>QN_MAX_LAYERS)
    {
	clog.error("Requested %d layers, maximum is %d.",
		   a_layers, QN_MAX_LAYERS);
    }
    // Initialize our layer sizes.
    for (i = 0; i<n_layers; i++)
    {
	if (a_layer_units!=NULL)
	    layer_units[i] = a_layer_units[i];
	else
	    layer_units[i] = 0;
    }

    if (a_mode==QN_READ)
    {
	clog.log(QN_LOG_PER_EPOCH, "Accessing Tnet weight file '%s' for reading.",
		 QN_FILE2NAME(a_stream));

	// TODO - use a_skip_sections.
	assert(a_skip_sections==0);

	scan_file();

	clog.log(QN_LOG_PER_EPOCH, "Weight file '%s' appears to have %lu layers.",
		 QN_FILE2NAME(stream), (unsigned long) n_layers);

	for (i=0; i<n_layers; i++)
	{
	    clog.log(QN_LOG_PER_EPOCH, "Layer %lu has %lu units.",
		     (unsigned long) i+1, (unsigned long) layer_units[i]);
	}

	// If constructor specifies number of layers, check file agress.
	if (a_layers!=0 && a_layers != n_layers)
	{
	    clog.error("Tnet Weight file constructor requested %lu layers, "
		       "file '%s' only had %lu layers.",
		       a_layers, QN_FILE2NAME(stream), n_layers);
	}
	// If constructor specifies size of layer, check file agrees.
	for (i=0; i<n_layers; i++)
	{
	    if (a_layer_units!=NULL && a_layer_units[i]!=0
		&& a_layer_units[i]!=layer_units[i])
	    {
		clog.error("Tnet Weight file constructor requested "
			   "layer %lu had %lu units, file '%s' had "
			   "%lu in the layer.",
			   (unsigned long) i+1,
			   (unsigned long) a_layer_units[i],
			   QN_FILE2NAME(stream),
			   (unsigned long) layer_units[i]);
	    }
	}
	// QN_WEIGHTS_UNKNOWN is used to signal start of file.
	cur_input_layer = 0;
	seek_layer_header();
    }
    else if (a_mode == QN_WRITE)
    {
	clog.log(QN_LOG_PER_EPOCH, "Accessing Tnet weight file "
		 "'%s' for writing.", QN_FILE2NAME(a_stream));
	for (i = 0; i< n_layers; i++)
	{
	    if (layer_units[i] == 0)
		clog.error("Cannot specify layer %d to have 0 units when writing.");
	}
	if (skip_sections != 0)
	    clog.error("Cannot skip sections when writing.");
	// QN_WEIGHTS_UNKNOWN is used to signal start of file.
	output_state = QN_WEIGHTS_UNKNOWN;
	// Write the header for the current, intial, weight block
	write_section_header();
    }
    else
    {
	clog.error("Unknown file access mode.");
    }
}

QN_MLPWeightFile_Tnet::~QN_MLPWeightFile_Tnet()
{
}


size_t
QN_MLPWeightFile_Tnet::num_layers()
{
    return n_layers;
}

size_t
QN_MLPWeightFile_Tnet::num_sections()
{
    return (n_layers-1)*2;
}

const char*
QN_MLPWeightFile_Tnet::layer_string(QN_OutputLayerType type)
{
    const char* str = NULL;

    switch(type)
    {
    case QN_LAYER_SIGMOID:
    case QN_LAYER_SIGMOID_XENTROPY:
	str =  "sigmoid";
	break;
    case QN_LAYER_SOFTMAX:
	str =  "softmax";
	break;
    case QN_LAYER_LINEAR:
	clog.error("Do not know header for TNet weight file LINEAR layer");
	break;
    case QN_LAYER_TANH:
	clog.error("Do not know header for TNet weight file TANH layer");
    default:
	clog.error("Uknknown output layer type writing TNet weight file");
    }
    return str;
}

void
QN_MLPWeightFile_Tnet::do_weights_header(size_t layer,
					 size_t next_state)
{
    size_t out = 0;		// Number of output units
    size_t in = 0;		// Number of input units
    int rc;

    // printf("starting layer %lu-%lu weights\n", (unsigned long) (layer-1), (unsigned long) layer);
    if (n_layers>layer)	// More layers to come
    {
	rc = fprintf(stream, "<sigmoid> %lu %lu\n",
		     output_outputs, output_outputs);
	if (rc<=0)
	{
	    clog.error("Error writing to TNet weight file '%s' - %s.",
		       QN_FILE2NAME(stream), strerror(errno));
	}
	out = layer_units[layer];
	in = layer_units[layer-1];
	output_state = next_state;
	output_outputs = out;
	output_inputs = in;
	output_row_count = out;
	output_col_count = in;
	rc = fprintf(stream, "<biasedlinearity> %lu %lu\nm %lu %lu\n",
		     out, in, out, in);
	if (rc<=0)
	{
	    clog.error("Error writing to TNet weight file '%s' - %s.",
		       QN_FILE2NAME(stream), strerror(errno));
	}
    }
    else			// Last layer
    {
	rc = fprintf(stream, "<%s> %lu %lu\n",
		     layer_string(out_layer_type),
		     output_outputs, output_outputs);
	if (rc<=0)
	{
	    clog.error("Error writing to TNet weight file '%s' - %s.",
		       QN_FILE2NAME(stream), strerror(errno));
	}
	output_state = QN_WEIGHTS_UNKNOWN;
    }
}

void
QN_MLPWeightFile_Tnet::do_bias_header(size_t layer,
				      size_t next_state)
{
    size_t out;
    int rc;

    // printf("starting layer %lu biases\n", (unsigned long) layer);
    out = layer_units[layer];
    // Not much of a header for biases - just the vector header.
    rc = fprintf(stream, "v %lu  ", out);
    if (rc<=0)
    {
	clog.error("Error writing to TNet weight file '%s' - %s.",
		   QN_FILE2NAME(stream), strerror(errno));
    }
    output_state = next_state;
    output_row_count = 1;
    output_col_count = layer_units[layer];
}


void
QN_MLPWeightFile_Tnet::write_section_header(void)
{
    size_t ret;
    size_t out = 0;		// Number of output units
    size_t in = 0;		// Number of input units
    int rc;			// Return code


    switch(output_state)
    {
    case QN_WEIGHTS_UNKNOWN:
     	out = layer_units[1];
	in = layer_units[0];
    	output_state = QN_LAYER12_WEIGHTS;
	output_outputs = out;
	output_inputs = in;
	output_row_count = out;
	output_col_count = in;
	rc = fprintf(stream, "<biasedlinearity> %lu %lu\nm %lu %lu\n",
		     out, in, out, in);
	if (rc<=0)
	{
	    clog.error("Error writing to TNet weight file '%s' - %s.",
		       QN_FILE2NAME(stream), strerror(errno));
	}
     	break;
    case QN_LAYER12_WEIGHTS:
	do_bias_header(1, QN_LAYER2_BIAS);
	break;
    case QN_LAYER2_BIAS:
	do_weights_header(2, QN_LAYER23_WEIGHTS);
	break;
    case QN_LAYER23_WEIGHTS:
	do_bias_header(2, QN_LAYER3_BIAS);
	break;
    case QN_LAYER3_BIAS:
	do_weights_header(3, QN_LAYER34_WEIGHTS);
	break;
    case QN_LAYER34_WEIGHTS:
	do_bias_header(3, QN_LAYER4_BIAS);
	break;
    case QN_LAYER4_BIAS:
	do_weights_header(4, QN_LAYER45_WEIGHTS);
	break;
    case QN_LAYER45_WEIGHTS:
	do_bias_header(4, QN_LAYER5_BIAS);
	break;
    case QN_LAYER5_BIAS:
	do_weights_header(5, QN_LAYER56_WEIGHTS);
	break;
    case QN_LAYER56_WEIGHTS:
	do_bias_header(5, QN_LAYER6_BIAS);
	break;
    case QN_LAYER6_BIAS:
	do_weights_header(6, QN_LAYER67_WEIGHTS);
	break;
    case QN_LAYER67_WEIGHTS:
	do_bias_header(6, QN_LAYER7_BIAS);
	break;
    case QN_LAYER7_BIAS:
	do_weights_header(7, QN_LAYER78_WEIGHTS);
	break;
    case QN_LAYER78_WEIGHTS:
	do_bias_header(7, QN_LAYER8_BIAS);
	break;
    case QN_LAYER8_BIAS:
	do_weights_header(8, QN_LAYER89_WEIGHTS);
	break;
    case QN_LAYER89_WEIGHTS:
	do_bias_header(8, QN_LAYER9_BIAS);
	break;
    case QN_LAYER9_BIAS:
	do_weights_header(9, QN_LAYER_UNKNOWN);
	break;
    }
}

size_t
QN_MLPWeightFile_Tnet::write(const float* buf, size_t count)
{
    size_t rv;
    int rc;
    size_t io_count;

    rv = count;			// Remember total count for later.
    if (mode != QN_WRITE)
    {
	clog.error("Tried to write readable Tnet weights file '%s'.",
		   QN_FILE2NAME(stream));
    }
    if (output_state==QN_WEIGHTS_UNKNOWN)
    {
 	clog.error("Tried to write beyond end of weight file '%s'.",
		   QN_FILE2NAME(stream));
    }
    io_count = output_inputs * (output_row_count-1) + output_col_count;
    // printf("output_state=%lu\n", (unsigned long) output_state);
    // printf("io_count=%lu\n", (unsigned long) io_count);
    // printf("count=%lu\n", (unsigned long) count);
    if (count > io_count)
    {
    	clog.error("Tried to write %lu values to '%s' when only %lu space in "
    		   "the current matrix.", count, QN_FILE2NAME(stream),io_count);
    }
    while (count > 0)
    {
	rc = fprintf(stream, "%g ", *buf);
	if (rc<=0)
	{
	    clog.error("Error writing to TNet weight file '%s' - %s.",
		       QN_FILE2NAME(stream), strerror(errno));
	}
	output_col_count--;
	if (output_col_count==0)
	{
	    output_row_count--;
	    output_col_count = output_inputs;
	    rc = fprintf(stream, "\n");
	    if (rc!=1)
	    {
		clog.error("Error writing to TNet weight file '%s' - %s.",
			   QN_FILE2NAME(stream), strerror(errno));
	    }
	}
	count--;
	buf++;
	io_count--;
    }
    if (io_count == 0)		// If we have done this matrix, on to next one.
	write_section_header();
    return rv;
}


void
QN_MLPWeightFile_Tnet::skip_bias_header()
{
    int ec;
    QN_Tnet_layer*  tnet_layer;	// Current tnet layer.
    size_t n_out;		// The number of outputs in the current layer.
    int c;
    size_t vec_size;		// Length of vector read in.

    tnet_layer = &(tnet_layers[cur_input_layer]);
    n_out = tnet_layer->n_output;
    input_count = n_out;
    input_biases = QN_TRUE;

    c = mygetc_eofbad();
    while (isspace(c))
    {
	if (c=='\n')
	    lineno++;
	c = mygetc_eofbad();
    }
    if (c!='v')
	clog.error("Error reading weight file '%s' - bad vector type in <biasedlinearity> during actual read at line %lu.", QN_FILE2NAME(stream), lineno);
    
    vec_size = mygetsize_eofbad();
    if (vec_size != n_out)
    {
	clog.error("Error reading weight file '%s' - bad vector header in <biasedlinearity> during actual read at line %lu.", QN_FILE2NAME(stream), lineno);
    }
}

void
QN_MLPWeightFile_Tnet::seek_layer_header()
{
    int ec;
    QN_Tnet_layer*  tnet_layer;	// Current tnet layer.
    size_t n_out;		// The number of outputs in the current layer.
    size_t n_in;		// The number of inputs in the current layer.

    cur_input_layer++;
    if (cur_input_layer>=n_layers)
    {
	cur_input_layer = 0;
	return;
    }

    tnet_layer = &(tnet_layers[cur_input_layer]);
    n_out = tnet_layer->n_output;
    n_in = tnet_layer->n_input;
    input_count = n_out * n_in;
    input_biases = QN_FALSE;
    ec = fsetpos(stream, &(tnet_layer->pos));
    lineno = tnet_layer->lineno;
    if (ec!=0)
    {
	clog.error("Error seeking to layer %lu in TNet weight file '%s',",
		   (unsigned long) cur_input_layer, QN_FILE2NAME(stream));
    }
}

enum QN_SectionSelector
QN_MLPWeightFile_Tnet::get_weighttype(int section)
{
    enum QN_SectionSelector rc = QN_WEIGHTS_UNKNOWN;

    switch(section)
    {
    case 0:
	rc =  QN_LAYER12_WEIGHTS;
	break;
    case 1:
	rc = QN_LAYER2_BIAS;
	break;
    case 2:
	if (n_layers>2)
	    rc =  QN_LAYER23_WEIGHTS;
	break;
    case 3:
	if (n_layers>2)
	    rc =  QN_LAYER3_BIAS;
	break;
    case 4:
	if (n_layers>3)
	    rc =  QN_LAYER34_WEIGHTS;
	break;
    case 5:
	if (n_layers>3)
	    rc =  QN_LAYER4_BIAS;
	break;
    case 6:
	if (n_layers>4)
	    rc =  QN_LAYER45_WEIGHTS;
	break;
    case 7:
	if (n_layers>4)
	    rc =  QN_LAYER5_BIAS;
	break;
    case 8:
	if (n_layers>5)
	    rc =  QN_LAYER56_WEIGHTS;
	break;
    case 9:
	if (n_layers>5)
	    rc =  QN_LAYER6_BIAS;
	break;
    case 10:
	if (n_layers>6)
	    rc =  QN_LAYER67_WEIGHTS;
	break;
    case 11:
	if (n_layers>6)
	    rc =  QN_LAYER7_BIAS;
	break;
    case 12:
	if (n_layers>7)
	    rc =  QN_LAYER78_WEIGHTS;
	break;
    case 13:
	if (n_layers>7)
	    rc =  QN_LAYER8_BIAS;
	break;
    case 14:
	if (n_layers>8)
	    rc =  QN_LAYER89_WEIGHTS;
	break;
    case 15:
	if (n_layers>8)
	    rc =  QN_LAYER9_BIAS;
	break;
    }
    if (rc==QN_WEIGHTS_UNKNOWN)
    {
	clog.error("Trying to get section %d weights when we only have "
		   "%lu layers.", section, (unsigned long) n_layers);
    }
    return rc;
}

size_t
QN_MLPWeightFile_Tnet::read(float* dest, size_t count)
{
    size_t res;
    size_t i;

    if (mode != QN_READ)
    {
	clog.error("Tried to read writeable Tnet weights file '%s'.",
		   QN_FILE2NAME(stream));
    }

    if (input_count==0)
    {
	if (input_biases)
	    seek_layer_header();
	else
	    skip_bias_header();
    }

    if (cur_input_layer==0)
    {
    	clog.error("Tried to read past end of Tnet weights file '%s'.",
    		   QN_FILE2NAME(stream));
    }
    if (input_count<count)
    {
	clog.error("Tried to read past end of section in Tnet weights "
		   " file '%s'.", QN_FILE2NAME(stream));
    }
    for (i=0; i<count; i++)
    {
	float f; 
	f = (float) mygetdouble_eofbad();
	// printf("read %f.\n", f);
	*dest++ = f;
    }
    input_count -= count;
    return count;
}


size_t
QN_MLPWeightFile_Tnet::size_layer(QN_LayerSelector layer)
{
    size_t res = QN_SIZET_BAD;

    switch(layer)
    {
    case QN_LAYER1:
	res = layer_units[0];
	break;
    case QN_LAYER2:
	res = layer_units[1];
	break;
    case QN_LAYER3:
	if (n_layers>2)
	    res = layer_units[2];
	break;
    case QN_LAYER4:
	if (n_layers>3)
	    res = layer_units[3];
	break;
    case QN_LAYER5:
	if (n_layers>4)
	    res = layer_units[4];
	break;
    case QN_LAYER6:
	if (n_layers>5)
	    res = layer_units[5];
	break;
    case QN_LAYER7:
	if (n_layers>6)
	    res = layer_units[6];
	break;
    case QN_LAYER8:
	if (n_layers>7)
	    res = layer_units[7];
	break;
    case QN_LAYER9:
	if (n_layers>8)
	    res = layer_units[8];
	break;
    default:
	res = 0;
	clog.error("size_layer: uknown layer number %lu.",
		   (unsigned long) layer);
	break;
    }
    if (res==QN_SIZET_BAD)
    {
	clog.error("size_layer: layer %lu requested, this net only "
		   "has %lu layers.",
		   (unsigned long) layer, (unsigned long) n_layers);
    }
    return res;
}

enum QN_WeightMaj
QN_MLPWeightFile_Tnet::get_weightmaj()
{
    // TODO - is this right?
    return QN_OUTPUTMAJOR;
}

enum QN_FileMode
QN_MLPWeightFile_Tnet::get_filemode()
{
    return mode;
}

int
QN_MLPWeightFile_Tnet::mygetc_eofbad()
{
    int c;

    c = getc(stream);
    if (c==EOF)
    {
	if (ferror(stream))
	{
	    clog.error("Error reading weight file '%s' - %s.", QN_FILE2NAME(stream), strerror(errno));
	}
	else
	{
	    clog.error("Error reading weight file '%s' - premature end of file reading character.", QN_FILE2NAME(stream));
	}
    }
    return c;
}

int
QN_MLPWeightFile_Tnet::mygetc_eofok()
{
    int c;

    c = getc(stream);
    if (c==EOF)
    {
	if (ferror(stream))
	{
	    clog.error("Error reading weight file '%s' - %s.", QN_FILE2NAME(stream), strerror(errno));
	}
    }
    return c;
}

void
QN_MLPWeightFile_Tnet::mygetpos(fpos_t* pos)
{
    int c;

    c = fgetpos(stream, pos);
    if (c!=0)
    {
	clog.error("Error reading weight file '%s' - %s.", QN_FILE2NAME(stream), strerror(errno));
    }
}

void
QN_MLPWeightFile_Tnet::myungetc(int c)
{
    int e;

    e = ungetc(c, stream);

    if (e==EOF)
	clog.error("Error pushing back input char reading weightfile '%s'.", QN_FILE2NAME(stream));
}

void
QN_MLPWeightFile_Tnet::myskipspace_eofbad(void)
{
    int c;

    c = mygetc_eofbad();
    while (isspace(c))
    {
	if (c=='\n')
	    lineno++;
	c = mygetc_eofbad();
    }
    myungetc(c);
}

unsigned long
QN_MLPWeightFile_Tnet::mygetsize_eofbad(void)
{
    int e;
    unsigned long size;

    myskipspace_eofbad();
    e = fscanf(stream, "%lu", &size);
    if (e!=1)
    {
	if (ferror(stream))
	{
	    clog.error("Error reading weight file '%s' - %s.", QN_FILE2NAME(stream), strerror(errno));
	}
	else if (feof(stream))
	{
	    clog.error("Error reading weight file '%s' - premature end of file reading size.", QN_FILE2NAME(stream));
	}
	else
	    clog.error("Error reading weight file '%s' - problem reading size at line %lu.", QN_FILE2NAME(stream), lineno);
    }
    return size;
}

double
QN_MLPWeightFile_Tnet::mygetdouble_eofbad(void)
{
    int e;
    double f;

    myskipspace_eofbad();
    e = fscanf(stream, "%lf", &f);
    if (e!=1)
    {
	if (ferror(stream))
	{
	    clog.error("Error reading weight file '%s' - %s.", QN_FILE2NAME(stream), strerror(errno));
	}
	else if (feof(stream))
	{
	    clog.error("Error reading weight file '%s' - premature end of file reading value.", QN_FILE2NAME(stream));
	}
	else
	    clog.error("Error reading weight file '%s' - problem reading floating point value at line %lu.", QN_FILE2NAME(stream), lineno);
    }
    return f;
}

void
QN_MLPWeightFile_Tnet::scan_file(void)
{
    int ec;			// Error code
    fpos_t pos;			// Current position
    int cnt;			// Count of characters for scanf
    size_t i;			// Local counter
    const unsigned int MAX_SECTNAME = 32; // Maximum size fo a section name.
    char sectname[MAX_SECTNAME+1]; // Name of current section.
    unsigned long n_out;	   // Number of outputs in layer.
    unsigned long n_in;		   // Number of inputs in layer.
    unsigned long mat_size1;	   // Matrix first size.
    unsigned long mat_size2;	   // Matrix second size.
    int c;			   // Read character.
    double d;			   // Temp var.
    
    long int cur_input_layer = 0;	// Start before first layer.

    lineno = 1;
    c = mygetc_eofok();
    while (!feof(stream) && (cur_input_layer<=QN_MLP_MAX_LAYERS))
    {
	// Skip over initial/trailing white space
	while (isspace(c))
	{
	    if (c=='\n')
		lineno++;
	    c = mygetc_eofok();
	}
	if (c==EOF)
	    break;
	if (c!='<')
	{
	    clog.error("Error reading weight file '%s' - garbage before section name at line %lu.", QN_FILE2NAME(stream), lineno);
	}
	cnt = 0;
	c = mygetc_eofbad();
	while(isalpha(c) && cnt<=MAX_SECTNAME)
	{
	    sectname[cnt++] = c;
	    c = mygetc_eofbad();
	}
	if (cnt>=MAX_SECTNAME)
	    clog.error("Error reading weight file '%s' - section name >%lu chars at line %lu.", QN_FILE2NAME(stream), (unsigned long) MAX_SECTNAME, lineno);
	else
	    sectname[cnt] = '\0';
	if (c!='>')
	{
	    clog.error("Error reading weight file '%s' - garbage after section name at line %lu.", QN_FILE2NAME(stream), lineno);
	}
	n_out = mygetsize_eofbad();
	n_in = mygetsize_eofbad();
	if (strcmp(sectname, "biasedlinearity")==0)
	{
	    fpos_t pos;

	    cur_input_layer++;
	    tnet_layers[cur_input_layer].n_output = n_out;
	    tnet_layers[cur_input_layer].n_input = n_in;
	    // By default linear layer.
	    tnet_layers[cur_input_layer].layer_type = QN_LAYER_LINEAR;
	    c = mygetc_eofbad();
	    while (isspace(c))
	    {
		if (c=='\n')
		    lineno++;
		c = mygetc_eofbad();
	    }
	    if (c!='m')
		clog.error("Error reading weight file '%s' - bad matrix type in <biasedlinearity> at line %lu.", QN_FILE2NAME(stream), lineno);
	    mat_size1 = mygetsize_eofbad();
	    mat_size2 = mygetsize_eofbad();
	    if ((mat_size1 != n_out) || (mat_size2 != n_in))
	    {
		clog.error("Error reading weight file '%s' - bad matrix header in <biasedlinearity> at line %lu.", QN_FILE2NAME(stream), lineno);
	    }
	    mygetpos(&pos);
	    tnet_layers[cur_input_layer].pos = pos;
	    tnet_layers[cur_input_layer].lineno = lineno;

	    for (i = 0; i<(n_out*n_in); i++)
		d = mygetdouble_eofbad();
	    c = mygetc_eofbad();
	    while (isspace(c))
	    {
		if (c=='\n')
		    lineno++;
		c = mygetc_eofbad();
	    }
	    if (c!='v')
		clog.error("Error reading weight file '%s' - bad vector type in <biasedlinearity>. at line %lu", QN_FILE2NAME(stream), lineno);
	    mat_size1 = mygetsize_eofbad();
	    if (mat_size1 != n_out)
	    {
		clog.error("Error reading weight file '%s' - bad vector header in <biasedlinearity> at line %lu.", QN_FILE2NAME(stream), lineno);
	    }
	    // Note this assumes a trailing linefeed otherwise it will break.
	    for (i = 0; i<n_out; i++)
		d = mygetdouble_eofbad();
	}
	else if (strcmp(sectname, "sigmoid")==0)
	{
	    if (n_out!=n_in)
	    {
		clog.error("Error reading weight file '%s' - bad sigmoid header at line %lu.", QN_FILE2NAME(stream), lineno);
	    }
	    tnet_layers[cur_input_layer].layer_type = QN_LAYER_SIGMOID;
	}
	else if (strcmp(sectname, "softmax")==0)
	{
	    if (n_out!=n_in)
		clog.error("Error reading weight file '%s' - bad softmax header at line %lu", QN_FILE2NAME(stream), lineno);
	    tnet_layers[cur_input_layer].layer_type = QN_LAYER_SOFTMAX;
	}
	else
	{
	    clog.error("Error reading weight file '%s' - unknown section name '%s' - at line %lu", QN_FILE2NAME(stream), sectname[cnt], lineno);
	}
	c = mygetc_eofok();
    } // Loop over sections.
    if (cur_input_layer>QN_MLP_MAX_LAYERS)
    {
	clog.error("Error reading weight file '%s' - too many layers.", QN_FILE2NAME(stream));
    }
    if (cur_input_layer==0)
    {
	clog.error("Error reading weight file '%s' - no layers found.", QN_FILE2NAME(stream));
    }
    
    // Set up the state.
    n_layers = cur_input_layer + 1;
    layer_units[0] = tnet_layers[1].n_input;
    for (i = 1; i<n_layers; i++)
    {
	layer_units[i] = tnet_layers[i].n_output;
	if (tnet_layers[i].n_input != layer_units[i-1])
	{
	    clog.error("Error reading weight file '%s' - inconsistent TNet matrix sizes at layer %lu.", QN_FILE2NAME(stream), (unsigned long) i+1);
	}
    }
    // Check that intermediate layers are sigmoids - this is hard coded in
    // QuickNet.
    if (n_layers>2)
    {
	for (i = 1; i<(n_layers-1); i++)
	{
	    if (tnet_layers[i].layer_type != QN_LAYER_SIGMOID)
	    {
		clog.error("Error reading weight file '%s' - quicknet only supports sigmoids in non-output layers.", QN_FILE2NAME(stream));
	    }
	}
    }
    out_layer_type = (QN_OutputLayerType) tnet_layers[n_layers-1].layer_type;
}


